/************************************************************************
 *   IRC - Internet Relay Chat, include/translate.h
 *   Copyright (C) 2001 Dmitry Agaphonov
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: translate.h,v 1.1.12.6 2002/12/05 01:48:57 rzhe Exp $ */

#if !defined(__TRANSLATE_H__)
#define __TRANSLATE_H__


typedef unsigned char cpatom;
typedef struct Codepage aCodepage;
typedef struct Translator aTranslator;
typedef struct ReadData aReadData;

extern int rehash_codepages();
extern void assign_codepage(aClient *cptr, aCodepage *codepage);
extern void unassign_codepage(aClient *cptr);
extern void change_codepage(aClient *cptr, aCodepage *codepage);
extern void report_codepages(aClient *cptr);
extern Link *makecpslist_bystring(char *str, char *buf, int bufsize);
extern void set_name(aClient *cptr, char *name);
extern void add_to_nickx_list(char *name);
extern void del_from_nickx_list(char *name);

#ifdef PROFILING
extern char *translate_frombase(aCodepage *codepage, char *msg, int len);
extern char *translate_tobase(const aCodepage *codepage, char *msg, int len);
extern void translate_frombase_inplace(const aCodepage *codepage, char *msg,
		int len);
extern void translate_tobase_inplace(const aCodepage *codepage, char *msg,
		int len);
extern int isvalid_4nicks(const aCodepage *codepage, cpatom a);
extern int isvalid_4chans(const aCodepage *codepage, cpatom a);
extern void reset_codepages_sendbufs();
extern int getstrflags(const aCodepage *codepage, const char *str);
#endif

extern aCodepage       *codepages, *base_codepage;
extern Link	       *b_lines, *p_lines;
extern int		blines_number;
extern char		protoctl_8bitinfo[];

struct Translator
{
	cpatom		from[CODEPAGE_HALFSIZE];
	cpatom		to[CODEPAGE_HALFSIZE];
};

struct ReadData
{
	char	       *alias_str;
	cpatom	       *codepage;
	cpatom	       *tolowertab;
	cpatom	       *touppertab;
	unsigned char  *flagstab;
	unsigned char  *identab;
};

#define CODEPAGE_BYTES		(CODEPAGE_SIZE * sizeof(cpatom))
#define CODEPAGE_HALFBYTES	(CODEPAGE_HALFSIZE * sizeof(cpatom))

/* Codepage flags */
#define CODEPAGE_ACTIVE		0x0001 /* Whether the codepage exists in the
					  latest configuration */
#define CODEPAGE_NOTMAXCLIENTS	0x0002 /* Indicates that a same active
					  codepage with greater number of
					  clients exists */
#define CODEPAGE_BASE		0x0004 /* Indicates base codepage */

/* Atom flags */
#define ATOM_VALID4NICKS	0x01
#define ATOM_VALID4CHANS	0x02
#define ATOM_LATIN		0x04
#define ATOM_NATIVE		0x08
#define ATOM_DIGIT		0x10
#define ATOM_8BIT		0x80 /* Reserved */
#define ATOM_MIXEDALPHA		(ATOM_LATIN | ATOM_NATIVE)
#define ATOM_ALPHANUMERIC	(ATOM_LATIN | ATOM_NATIVE | ATOM_DIGIT)

/* String flags */
#define STR_HAS8BIT		0x0080
#define STR_LATIN		ATOM_LATIN
#define STR_NATIVE		ATOM_NATIVE
#define STR_DIGIT		ATOM_DIGIT
#define STR_MIXEDALPHA		ATOM_MIXEDALPHA
#define STR_MIXEDALPHAWORDS	0x0100
#define STR_ALL			(STR_HAS8BIT | STR_MIXEDALPHA | STR_DIGIT | \
					STR_MIXEDALPHAWORDS)

struct Codepage
{
	char		name[MAX_CODEPAGE_NAME_LENGTH + 1];
	/* Just for having more information in stats, etc. */
	char		filename[MAX_CODEPAGE_FILENAME_LENGTH + 1];
	/* Number of clients using this codepage */
	Link	       *aliases;
	int		clients;
	int		maxclients;
	int		flags;
	
	/* Translator from the absolute codepage to this one */
	cpatom		codepage[CODEPAGE_SIZE];
	cpatom		tolowertab[CODEPAGE_HALFSIZE];
	cpatom		touppertab[CODEPAGE_HALFSIZE];
	unsigned char	flagstab[CODEPAGE_HALFSIZE];
	unsigned char	identab[CODEPAGE_HALFSIZE];
	/* Translator from and to the base codepage */
	aTranslator    *translator;
	char		sendbuf[TRANSLATE_SENDBUFFER_SIZE];

	/*
	 * Still not decided whether to keep accosiated aconfs here.
	 * Now used for accosiated port numbers list.
	 */
	Link	       *ports;

	/* Infos */
	char		str_flags[5], str_ports[64];

	/* Auxiliary data used while rehashing codepages */
	aReadData	readdata;
	
	aCodepage      *prev;
	aCodepage      *next;
};

#define	codepage_fromabs	codepage
#define	codepage_toabs		codepage + CODEPAGE_HALFSIZE


#ifndef PROFILING
#define __func(type) static inline type
#include "translate.inc"
#endif


#endif /* !__TRANSLATE_H__ */
