/************************************************************************
 *   IRC - Internet Relay Chat, include/translate.inc
 *   Copyright (C) 2002 Dmitry Agaphonov
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: translate.inc,v 1.1.8.4 2002/12/02 21:14:16 rzhe Exp $ */


__func(char) *translate_frombase(aCodepage *codepage, char *msg, int len)
{
	cpatom *from;
	int	i;
	char   *buf;

	if (!codepage || !codepage->translator)
		return msg;

	if (!(*codepage->sendbuf))
	{
		from = codepage->translator->from;
		buf = codepage->sendbuf;

		/*
		 * Not checking for len > TRANSLATE_BUFFER_SIZE to not waste
		 * CPU.
		 * Just assume that this is a macro and do not call this
		 * function with len is greater than TRANSLATE_BUFFER_SIZE.
		 * -rzhe
		 */
		for (i = 0; i < len; i++)
			*buf++ = from[(unsigned char)*msg++];
		*buf = 0;
	}

	return codepage->sendbuf;
}

__func(char) *translate_tobase(const aCodepage *codepage, char *msg, int len)
{
	static char translate_readbuf[TRANSLATE_READBUFFER_SIZE];

	cpatom *to;
	int	i;
	char   *buf = translate_readbuf;

	if (!codepage || !codepage->translator)
		return msg;

	to = codepage->translator->to;

	/*
	 * Not checking for len > TRANSLATE_BUFFER_SIZE to not waste CPU.
	 * Just assume that this is a macro and do not call this function
	 * with len is greater than TRANSLATE_BUFFER_SIZE.
	 * -rzhe
	 */
	for (i = 0; i < len; i++)
		*buf++ = to[(unsigned char)*msg++];
	*buf = 0;

	return translate_readbuf;
}

__func(void) translate_frombase_inplace(const aCodepage *codepage, char *msg,
		int len)
{
	cpatom *from;
	int	i;

	if (!codepage || !codepage->translator)
		return;

	from = codepage->translator->from;

	/*
	 * Not checking for len > TRANSLATE_BUFFER_SIZE to not waste CPU.
	 * Just assume that this is a macro and do not call this function
	 * with len is greater than TRANSLATE_BUFFER_SIZE.
	 * -rzhe
	 */
	for (i = 0; i < len; i++)
	{
		*msg = from[(unsigned char)*msg];
		msg++;
	}
}

__func(void) translate_tobase_inplace(const aCodepage *codepage, char *msg,
		int len)
{
	cpatom *to;
	int	i;

	if (!codepage || !codepage->translator)
		return;

	to = codepage->translator->to;

	/*
	 * Not checking for len > TRANSLATE_BUFFER_SIZE to not waste CPU.
	 * Just assume that this is a macro and do not call this function
	 * with len is greater than TRANSLATE_BUFFER_SIZE.
	 * -rzhe
	 */
	for (i = 0; i < len; i++)
	{
		*msg = to[(unsigned char)*msg];
		msg++;
	}
}

__func(int) isvalid_4nicks(const aCodepage *codepage, cpatom a)
{
	/*
	 * Not checking for !codepage to not waste CPU (this function is
	 * "called" in a loop).
	 * Just assume that this is a macro and do not call this function
	 * with codepage == NULL.
	 * -rzhe
	 */
	return codepage->flagstab[a] & ATOM_VALID4NICKS;
}

__func(int) isvalid_4chans(const aCodepage *codepage, cpatom a)
{
	/*
	 * Not checking for !codepage to not waste CPU (this function is
	 * "called" in a loop).
	 * Just assume that this is a macro and do not call this function
	 * with codepage == NULL.
	 * -rzhe
	 */
	return codepage->flagstab[a] & ATOM_VALID4CHANS;
}

__func(void) reset_codepages_sendbufs()
{
	aCodepage *cp;

	cp = codepages;
	while (cp)
	{
		*cp->sendbuf = 0;
		cp = cp->next;
	}
}

__func(int) getstrflags(const aCodepage *codepage, const char *str)
{
	const unsigned char *flagstab = codepage ? codepage->flagstab : NULL;
	unsigned char        f, wordflags = 0;
	int                  flags = 0;

	if (flagstab)
		for (; *str && ((flags & STR_ALL) != STR_ALL); str++)
		{
			f = flagstab[(cpatom)*str];

			flags |= (*str & 0x80) | f;

			if (f & ATOM_ALPHANUMERIC)
				wordflags |= f;
			else
				wordflags = 0;
			if ((wordflags & ATOM_MIXEDALPHA) == ATOM_MIXEDALPHA)
				flags |= STR_MIXEDALPHAWORDS;
		}
	else
		for (; *str && !flags; flags = *str++ & 0x80) ;

	return flags;
}


#undef __func
