/************************************************************************
 *   IRC - Internet Relay Chat, src/translate.c
 *   Copyright (C) 2001 Dmitry Agaphonov
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: translate.c,v 1.1.12.9 2003/01/07 11:33:03 rzhe Exp $ */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include "struct.h"
#include "config.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "h.h"
#include "translate.h"


aCodepage *codepages = NULL;
aCodepage *base_codepage = NULL;

Link *b_lines = NULL;
Link *p_lines = NULL;
int blines_number = 0;

char protoctl_8bitinfo[256];

struct protoctl_info
{
	char *cm;
	char *cps;
	char *_8bit, *ma, *maw;
	char *_8bnci, *_8bcnci;
	char *iden;
};

static int base_cp_changed = 1;	/* Indicates that base codepage content (not
				   only pointer!) has been changed */
static int base_cp_set = 0;	/* FIXME:
				   Should be removed when changing base
				   codepage support is added. -rzhe */

static Link *nickx = NULL;


#if defined(DEBUG_CODEPAGES)
static int dt_helper(char *fname, aCodepage *codepage, int from)
{
	int	fd;
	cpatom *data;
	size_t	size;

	if (codepage->translator)
	{
		data = from ? codepage->translator->from :
			codepage->translator->to;
		size = CODEPAGE_HALFBYTES;
	}
	else
	{
		data = "0";
		size = 1;
	}
	
	if (
		(fd = open(fname, O_CREAT | O_WRONLY, 0644)) == -1 ||
		write(fd, data, size) != size ||
		close(fd)
	)
	{
		/* Do something */
		return -1;
	}

	return 0;
}

static int dump_translator(aCodepage *codepage)
{
	char	fname[MAXPATHLEN];
	int	rc;

	ircsnprintf(fname, MAXPATHLEN, "%s/%s[BASE]_2_%s", dpath,
		base_codepage ? base_codepage->name : "", codepage->name);

	rc = dt_helper(fname, codepage, YES);

	ircsnprintf(fname, MAXPATHLEN, "%s/%s_2_%s[BASE]", dpath,
		codepage->name, base_codepage ? base_codepage->name : "");

	return dt_helper(fname, codepage, NO) || rc;
}
#endif /* DEBUG_CODEPAGES */

inline int inc_clients(aCodepage *cp)
{
	if (cp && ++cp->clients > cp->maxclients)
		cp->maxclients = cp->clients;

	return cp ? cp->clients : 0;
}

static void set_basecodepage(aCodepage *cp)
{
	if (base_codepage)
	{
		base_codepage->flags &= ~CODEPAGE_BASE;
		base_codepage->clients--;
	}

	/*
	 * Keep always at least one link to prevent deleting the base codepage
	 * even if the last client using it is destroyed.
	 * -rzhe
	 */
	if ((base_codepage = cp))
	{
		cp->flags |= CODEPAGE_BASE;
		cp->clients++;
#ifdef _8BITNAMES_CASEINSENSITIVE
		tolowertab = cp->tolowertab;
		touppertab = cp->touppertab;
#endif
#ifdef NICKNAMES_IDENTITY
		identab = cp->identab;
#endif
	}
	else
	{
#ifdef _8BITNAMES_CASEINSENSITIVE
		tolowertab = standard_tolowertab;
		touppertab = standard_touppertab;
#endif
#ifdef NICKNAMES_IDENTITY
		identab = standard_identab;
#endif
	}
}

#define	CPFILE_BYTES	(CODEPAGE_BYTES + CODEPAGE_HALFBYTES * 4)

#define	CP_OFFSET	(0)
#define	LOWER_OFFSET	(CODEPAGE_BYTES)
#define	UPPER_OFFSET	(CODEPAGE_BYTES + CODEPAGE_HALFBYTES)
#define	FLAGS_OFFSET	(CODEPAGE_BYTES + CODEPAGE_HALFBYTES * 2)
#define	IDEN_OFFSET	(CODEPAGE_BYTES + CODEPAGE_HALFBYTES * 3)

static int update_codepages()
{
	aCodepage      *cp, *t, *base_cp = NULL;
	aReadData      *rd;
	Link	       *bl, *lp;
	char	       *name, *filename, cpfnamebuf[MAXPATHLEN];
	char           *alias, *alias_str;
	unsigned char  *codepages_buffer, *buf;
	int		fd, is_base;

	cp = codepages;
	while (cp)
	{
		cp->flags &= ~CODEPAGE_ACTIVE & ~CODEPAGE_NOTMAXCLIENTS;

		rd = &cp->readdata;
		rd->alias_str = cp->aliases ? cp->aliases->value.cp : NULL;
		rd->codepage = cp->codepage;
		rd->tolowertab = cp->tolowertab;
		rd->touppertab = cp->touppertab;
		rd->flagstab = cp->flagstab;
		rd->identab = cp->identab;
		
		cp = cp->next;
	}
	
	if (!b_lines)
		return 0;

	buf = codepages_buffer = (unsigned char *)
		MyMalloc(CPFILE_BYTES * blines_number);

	bl = b_lines;
	while (bl)
	{
		ircsnprintf(cpfnamebuf, MAXPATHLEN, "%s/%s/%s", dpath, CPDIR,
			bl->value.aconf->passwd);

		if (
			(fd = open(cpfnamebuf, O_RDONLY)) == -1 ||
			read(fd, buf, CPFILE_BYTES) != CPFILE_BYTES ||
			close(fd)
		)
		{
			sendto_ops("Can't open, read or close %s file codepage could be missing! "
				"Codepages was not rehashed or initialized!",
				cpfnamebuf);
#if defined(USE_SYSLOG)
			syslog(LOG_WARNING, "Can't open, read or close %s file codepage could be missing! "
				"Codepages was not rehashed or initialized!",
				cpfnamebuf);
#endif
			MyFree(codepages_buffer);
			
			return -1;
		}

		buf += CPFILE_BYTES;
		bl = bl->next;
	}

	buf = codepages_buffer;
	while ((bl = b_lines))
	{
		name = bl->value.aconf->host;
		filename = bl->value.aconf->passwd;
		alias_str = !BadPtr(bl->value.aconf->name) ?
			bl->value.aconf->name : NULL;
		bl->flags = NO;
		is_base = bl->value.aconf->port & CODEPAGE_BASE;
		
		cp = codepages;
		while (cp)
		{
			rd = &cp->readdata;
			if (strcasecmp(name, cp->name))
			{
				if (!memcmp(buf + CP_OFFSET, rd->codepage, CODEPAGE_BYTES))
				{
					strncpyzt(cp->name, name, MAX_CODEPAGE_NAME_LENGTH);
					strncpyzt(cp->filename, filename, MAX_CODEPAGE_FILENAME_LENGTH);
					rd->alias_str = alias_str;
					rd->tolowertab = buf + LOWER_OFFSET;
					rd->touppertab = buf + UPPER_OFFSET;
					rd->flagstab = buf + FLAGS_OFFSET;
					rd->identab = buf + IDEN_OFFSET;

					cp->flags |= CODEPAGE_ACTIVE;

					bl->flags = YES;

					if (is_base)
						base_cp = cp;
				}
			}
			else
			{
				if (!memcmp(buf + CP_OFFSET, rd->codepage, CODEPAGE_BYTES))
				{
					strncpyzt(cp->filename, filename, MAX_CODEPAGE_FILENAME_LENGTH);
					rd->alias_str = alias_str;
					rd->tolowertab = buf + LOWER_OFFSET;
					rd->touppertab = buf + UPPER_OFFSET;
					rd->flagstab = buf + FLAGS_OFFSET;
					rd->identab = buf + IDEN_OFFSET;

					cp->flags |= CODEPAGE_ACTIVE;
					
					bl->flags = YES;

					if (is_base)
						base_cp = cp;
				}
				else
				{
					if (!cp->clients)
					{
						strncpyzt(cp->filename, filename, MAX_CODEPAGE_FILENAME_LENGTH);
						rd->alias_str = alias_str;
						rd->codepage = buf + CP_OFFSET;
						rd->tolowertab = buf + LOWER_OFFSET;
						rd->touppertab = buf + UPPER_OFFSET;
						rd->flagstab = buf + FLAGS_OFFSET;
						rd->identab = buf + IDEN_OFFSET;

						cp->flags |= CODEPAGE_ACTIVE;

						bl->flags = YES;

						if (is_base)
							base_cp = cp;
					}
					else
					{
						strncpyzt(cp->name, UNKNOWN_CODEPAGE_NAME, MAX_CODEPAGE_NAME_LENGTH);

						cp->flags &= ~CODEPAGE_ACTIVE;
					}
				}
			}
			
			cp = cp->next;
		}

		if (!bl->flags)
		{
			cp = make_codepage();

			strncpyzt(cp->name, name, MAX_CODEPAGE_NAME_LENGTH);
			strncpyzt(cp->filename, filename, MAX_CODEPAGE_FILENAME_LENGTH);
			cp->aliases = NULL;

			cp->clients = 0;
			cp->maxclients = 0;
			cp->flags = CODEPAGE_ACTIVE;

			cp->translator = NULL;
			*cp->sendbuf = 0;

			cp->ports = NULL;

			rd = &cp->readdata;
			rd->alias_str = alias_str;
			rd->codepage = buf + CP_OFFSET;
			rd->tolowertab = buf + LOWER_OFFSET;
			rd->touppertab = buf + UPPER_OFFSET;
			rd->flagstab = buf + FLAGS_OFFSET;
			rd->identab = buf + IDEN_OFFSET;

			cp->prev = NULL;
			cp->next = codepages;
			if (codepages)
				codepages->prev = cp;
			codepages = cp;

			if (is_base)
				base_cp = cp;
		}

		buf += CPFILE_BYTES;
		
		b_lines = bl->next;
		free_link(bl);
	}

	blines_number = 0;

	cp = codepages;
	while (cp)
	{
		t = cp->next;
		
		if (!(cp->clients || cp->flags & CODEPAGE_ACTIVE))
			remove_codepage(cp);
		else
		{
			rd = &cp->readdata;

			if (cp->aliases)
			{
				/*
				 * TODO:
				 * Optimize these two calls of strcasecmp() and
				 * strcmp() in one.
				 * -rzhe
				 */
				if (rd->alias_str ? strcasecmp(rd->alias_str,
						cp->aliases->value.cp) : 1)
				{
					remove_link_list(cp->aliases);
					cp->aliases = NULL;
				}
				else if (strcmp(rd->alias_str,
							cp->aliases->value.cp))
					strcpy(cp->aliases->value.cp,
							rd->alias_str);
			}

			if (rd->alias_str && !cp->aliases)
			{
				DupString(alias_str, rd->alias_str);
				for (alias = strtok(alias_str, ","); alias;
						alias = strtok(NULL, ","))
				{
					lp = make_link();
					DupString(lp->value.cp, alias);
					lp->next = cp->aliases;
					cp->aliases = lp;
				}
				MyFree(alias_str);

				lp = make_link();
				DupString(lp->value.cp, rd->alias_str);
				lp->next = cp->aliases;
				cp->aliases = lp;
			}

			if (rd->codepage != cp->codepage)
				memcpy(cp->codepage, rd->codepage, CODEPAGE_BYTES);

			/*
			 * A new codepage is being created.
			 * -rzhe
			 */
			if (rd->tolowertab != cp->tolowertab)
			{
				memcpy(cp->tolowertab, rd->tolowertab, CODEPAGE_HALFBYTES);
				memcpy(cp->touppertab, rd->touppertab, CODEPAGE_HALFBYTES);
				memcpy(cp->flagstab, rd->flagstab, CODEPAGE_HALFBYTES);
				memcpy(cp->identab, rd->identab, CODEPAGE_HALFSIZE);
			}
		}

		cp = t;
	}

	MyFree(codepages_buffer);

	/*
	 * FIXME:
	 * The following two lines should be replaced with the commented code
	 * below when changing base codepage support is added.
	 * -rzhe
	 */
	if (!base_cp_set)
	{
		base_cp_set = 1;
/*	if (base_codepage != base_cp)
	{
		base_cp_changed = !base_codepage || !base_cp ||
			memcmp(base_codepage->codepage, base_cp->codepage,
				CODEPAGE_BYTES);
*/
		set_basecodepage(base_cp);
	}
	
	return 0;
}

static void update_ports_codepages()
{
	Link	       *pl, *t, *lst;
	int		maxclients, cp_clients, basecpport_exists = 0;
	aClient	       *cptr;
	aCodepage      **_ptr_found_cp, *cp = NULL, *winner, *loser;

	pl = p_lines;
	while (pl)
	{
		pl->flags = 0;
		/*
		 * Choose the codepage with maximum clients number to make the
		 * server getting rid of duplicate codepages faster.
		 * -rzhe
		 */
		maxclients = -1;
		cp = codepages;
		while (cp)
		{
			if (
				!strcasecmp(pl->value.aconf->name, cp->name) &&
				cp->flags & CODEPAGE_ACTIVE
			)
			{
				_ptr_found_cp = (aCodepage**)&pl->flags;

				cp_clients = cp->clients;
				if (cp->flags & CODEPAGE_BASE)
					cp_clients--;
				
				/*
				 * Should '>=' be used in the comparison below?
				 * Putting '>=' leads to doing more operations
				 * here, but it also moves an active codepage
				 * closer to the beginning of the inverted
				 * p_lines list.
				 * -rzhe
				 */
				if (cp_clients > maxclients)
				{
					winner = cp;
					loser = *_ptr_found_cp;
				}
				else
				{
					winner = *_ptr_found_cp;
					loser = cp;
				}

				if (loser)
				{
					/*
					 * Make the base codepage having most of
					 * clients among the same codepages.
					 * -rzhe
					 */
					if (loser->flags & CODEPAGE_BASE)
						set_basecodepage(winner);

					if (loser->clients)
					{
						loser->flags |= CODEPAGE_NOTMAXCLIENTS;
						loser->flags &= ~CODEPAGE_ACTIVE;
					}
					else
						remove_codepage(loser);
				}

				if ((*_ptr_found_cp = winner))
				{
					maxclients = winner->clients;
					if (winner->flags & CODEPAGE_BASE)
						maxclients--;
				}
			}

			cp = cp->next;
		}

		pl = pl->next;
	}

#ifdef WARN_NO_BASE_CODEPAGE
	/*
	 * Do not spam operators if there is no codepages at all.
	 * -rzhe
	 */
	if (!base_codepage && codepages)
	{
		sendto_ops("No base codepage specified! "
			"Absolute codepage is used as a base. "
			"Highly not recommended!");
# if defined(USE_SYSLOG)
		syslog(LOG_WARNING, "No base codepage specified! "
			"Absolute codepage is used as a base. "
			"Highly not recommended!");
# endif
	}
#endif /* WARN_NO_BASE_CODEPAGE */
	
	cp = codepages;
	while (cp)
	{
		/*
		 * No set_basecodepage() calls expected now.
		 * -rzhe
		 */
		if (cp->clients > cp->maxclients)
			cp->maxclients = cp->clients;
		
		while (cp->ports)
		{
			t = cp->ports;
			cp->ports = t->next;
			free_link(t);
		}
		
		cp = cp->next;
	}

	/*
	 * FIXME:
	 * Fix the listeners list cycle bug.
	 * -rzhe
	 */
	for (lst = listeners; lst; lst = lst->next)
	{
		if (!(cptr = lst->value.cptr))
			continue;

		pl = p_lines;
		while (pl)
		{
			if (cptr->port == pl->value.aconf->port)
			{
				cp = (aCodepage*)pl->flags;
				if (cptr->codepage != cp)
				{
					if (cptr->codepage)
						cptr->codepage->clients--;

					if ((cptr->codepage = cp))
						(void) inc_clients(cp);
				}

				if (cp)
				{
					t = make_link();
					t->flags = cptr->port;
					t->next = cp->ports;
					cp->ports = t;

					if (cp->flags & CODEPAGE_BASE)
						basecpport_exists = 1;
				}

				break;
			}

			pl = pl->next;
		}

		/*
		 * There is no codepage associated with cptr->port.
		 * -rzhe
		 */
		if (!pl && cptr->codepage)
		{
			cptr->codepage->clients--;
			cptr->codepage = NULL;
		}
	}
	
	if (codepages && base_codepage && !basecpport_exists)
	{
		sendto_ops("There is no listening ports with base codepage! "
			"Highly not recommended!");
#if defined(USE_SYSLOG)
		syslog(LOG_WARNING, "There is no listening ports with base codepage! "
			"Highly not recommended!");
#endif
	}
	
	while ((pl = p_lines))
	{
		p_lines = pl->next;
		free_link(pl);
	}
}

static void update_translators()
{
	cpatom	       *base_fromabs, *fromabs;
	cpatom	       *t_iter, *f_iter, *ta_iter, *bta_iter;
	int		i;
	aTranslator    *translator;
	aCodepage      *cp;

	if (!codepages)
		return;

	base_fromabs = base_codepage ? base_codepage->codepage_fromabs : NULL;

	cp = codepages;
	while (cp)
	{
		if (!base_cp_changed && cp->translator)
		{
			cp = cp->next;
			continue;
		}
		
		if (cp->translator)
			free_translator(cp->translator);
		
		if (
			cp == base_codepage || (base_codepage &&
			!memcmp(cp->codepage, base_codepage->codepage,
				CODEPAGE_BYTES))
		)
			translator = NULL;
		else
		{
			fromabs = cp->codepage_fromabs;

			translator = make_translator();

			if (base_codepage)
			{
				t_iter = translator->to;
				f_iter = translator->from;
				ta_iter = cp->codepage_toabs;
				bta_iter = base_codepage->codepage_toabs;

				for (i = 0; i < CODEPAGE_HALFSIZE; i++)
				{
					*t_iter++ = base_fromabs[*ta_iter++];
					*f_iter++ = fromabs[*bta_iter++];
				}
			}
			else
			{
				memcpy(translator->from, cp->codepage_fromabs, CODEPAGE_HALFBYTES);
				memcpy(translator->to, cp->codepage_toabs, CODEPAGE_HALFBYTES);
			}
		}

		cp->translator = translator;
#if defined(DEBUG_CODEPAGES)
		dump_translator(cp);
#endif
		cp = cp->next;
	}
	
}

static aCodepage *findcodepage_byname(char *name)
{
	aCodepage *cp = codepages;
	Link     *link;
	
	while (cp)
	{
		if (cp->flags & CODEPAGE_ACTIVE)
		{
			if (!strcasecmp(cp->name, name))
				break;
			if (cp->aliases)
			{
				link = cp->aliases->next;
				while (link)
				{
					if (!strcasecmp(link->value.cp, name))
						break;
					link = link->next;
				}
				if (link)
					break;
			}
		}

		cp = cp->next;
	}

	return cp;
}

Link *makecpslist_bystring(char *str, char *buf, int bufsize)
{
	aCodepage      *cp;
	Link	       *newlink, *head, *tail;
	char	       *name;

	if (buf)
		*buf = '\0';

	head = tail = NULL;
	for (name = strtok(str, ","); name; name = strtok(NULL, ","))
		if ((cp = findcodepage_byname(name)))
		{
			newlink = make_link();
			newlink->value.codepage = cp;
			newlink->flags = 0;
			if (tail)
			{
				tail->next = newlink;
				tail = newlink;
			}
			else
				head = tail = newlink;

			if (buf)
			{
				if (*buf)
					strlcat(buf, ",", bufsize);
				if (cp)
					strlcat(buf, cp->name, bufsize);
			}
		}

	return head;
}

static void update_one_cps(Link *cps)
{
	char buf[64];

	if (cps)
	{
		cps->next = makecpslist_bystring(cps->value.cp, buf,
			sizeof(buf));

		MyFree(cps->value.cp);
		DupString(cps->value.cp, buf);
	}
}

static void update_classes_cps()
{
	aClass	       *cltmp;

	for (cltmp = FirstClass(); cltmp; cltmp = NextClass(cltmp))
		if (IsSvrClass(cltmp))
		{
			update_one_cps(XSvrCcps(cltmp));
			update_one_cps(XSvrNcps(cltmp));
		}
}

static void update_codepages_info()
{
	aCodepage      *cp;
	Link	       *p;
	char		padd[16], *f;
	int		i;

	cp = codepages;
	while (cp)
	{
		if (cp->flags)
		{
			f = cp->str_flags;
			if (cp->flags & CODEPAGE_ACTIVE)
				*f++ = 'A';
			if (cp->flags & CODEPAGE_NOTMAXCLIENTS)
				*f++ = 'N';
			if (cp->flags & CODEPAGE_BASE)
				*f++ = 'B';
			*f = 0;
		}
		else
			strcpy(cp->str_flags, "NONE");

		if (cp->ports)
		{
			*cp->str_ports = 0;
			i = 0;
			p = cp->ports;
			while (p)
			{
				i += sprintf(padd, "%ld", p->flags);
				/*
				 * Rude, but having so many ports is crazy.
				 * -rzhe
				 */
				if (i > sizeof(cp->str_ports) - 2)
					break;
				strcat(cp->str_ports, padd);

				if ((p = p->next))
				{
					strcat(cp->str_ports, ",");
					i++;
				}
			}
			/*
			 * Seems that cp->str_ports[255] = 0 is not needed.
			 * -rzhe
			 */
		}
		else
			strcpy(cp->str_ports, "NONE");

		cp = cp->next;
	}
}

static void fill_protoctl_info(struct protoctl_info *pi)
{
	aCodepage *cp;

	if (!base_codepage)
	{
		pi->cm = "CASEMAPPING=ascii ";
		pi->cps = pi->_8bit = pi->ma = pi->maw = pi->_8bnci =
			pi->_8bcnci = pi->iden = "";

		return;
	}

	pi->cm = "";

	pi->cps = "";
	for (cp = codepages; cp; cp = cp->next)
		if (cp->flags & CODEPAGE_ACTIVE)
		{
			pi->cps = "CODEPAGES ";
			break;
		}

#ifdef NO_8BIT_NICKNAMES
	pi->_8bit = "8BNICKS=NO ";
	pi->ma = pi->maw = "";
#else
	pi->_8bit = "8BNICKS=YES ";
# ifdef NO_MIXEDALPHA_NICKNAMES
	pi->ma = "MANICKS=NO ";
	pi->maw = "";
# else
	pi->ma = "MANICKS=YES ";
#  ifdef NO_MIXEDALPHAWORDS_NICKNAMES
	pi->maw = "MAWNICKS=NO ";
#  else
	pi->maw = "MAWNICKS=YES ";
#  endif
# endif /* NO_MIXEDALPHA_NICKNAMES */
#endif /* NO_8BIT_NICKNAMES */

#ifdef _8BITNAMES_CASEINSENSITIVE
	pi->_8bnci = "8BNCI=YES ";
# ifdef _8BITCHANNELNAMES_CASEINSENSITIVE
	pi->_8bcnci = "8BCNCI=YES ";
# else
	pi->_8bcnci = "8BCNCI=NO ";
# endif
#else
	pi->_8bnci = "8BNCI=NO ";
	pi->_8bcnci = "";
	pi->cm = "CASEMAPPING=ascii ";
#endif /* _8BITNAMES_CASEINSENSITIVE */

#ifdef NICKNAMES_IDENTITY
	pi->iden = "NICKIDEN ";
#else
	pi->iden = "";
#endif
}

static void update_protoctl_info()
{
	struct protoctl_info pi;

	fill_protoctl_info(&pi);

	ircsnprintf(protoctl_8bitinfo, sizeof(protoctl_8bitinfo),
		"%s%s%s%s%s%s%s%s", pi.cm, pi.cps, pi._8bit, pi.ma, pi.maw,
		pi._8bnci, pi._8bcnci, pi.iden);
}

int rehash_codepages()
{
	aCodepage      *cp, *t;
	int		rc;

	if ((rc = update_codepages()) < 0)
		return rc;
	
	update_ports_codepages();

	/*
	 * Active codepages with no clients should be kept if there is no
	 * the same active codepage which has clients.
	 * -rzhe
	 */
	cp = codepages;
	while (cp)
	{
		t = cp;
		cp = cp->next;
		/*
		 * TODO:
		 * Remove checking flags for CODEPAGE_NOTMAXCLIENTS bit set and
		 * CODEPAGE_NOTMAXCLIENTS flag concept by itself when sure that
		 * this is obsolete.
		 * -rzhe
		 */
		if (!t->clients && (t->flags & CODEPAGE_NOTMAXCLIENTS ||
					!(t->flags & CODEPAGE_ACTIVE)))
			remove_codepage(t);
	}

	update_translators();

	update_classes_cps();

	update_codepages_info();

	update_protoctl_info();

	return rc;
}

void assign_codepage(aClient *cptr, aCodepage *codepage)
{
	if (cptr->codepage)
	{
		sendto_ops("Client (%s) to assign a codepage (%s) already has a codepage (%s)! "
			"Please report this issue to ircd-RU! coding team!",
			cptr->name, codepage->name, cptr->codepage->name);
#if defined(USE_SYSLOG)
		syslog(LOG_WARNING, "Client (%s) to assign a codepage (%s) already has a codepage (%s)! "
			"Please report this issue to ircd-RU! coding team!",
			cptr->name, codepage->name, cptr->codepage->name);
#endif

		if (cptr->codepage == codepage)
			return;
		else
			unassign_codepage(cptr);
	}

	(void) inc_clients(cptr->codepage = codepage);
}

void unassign_codepage(aClient *cptr)
{
	aCodepage *codepage = cptr->codepage;
	
	if (!codepage)
		return;

	if (!(--codepage->clients || codepage->flags & CODEPAGE_ACTIVE))
		remove_codepage(codepage);

	cptr->codepage = NULL;
}

void change_codepage(aClient *cptr, aCodepage *codepage)
{
	if (cptr && cptr->codepage != codepage)
	{
		unassign_codepage(cptr);
		assign_codepage(cptr, codepage);
	}
}

static int find_free_nickx()
{
	int   suf;
	Link *lp;

	suf = 0;
	lp = nickx;
	while (lp && suf <= CHANGECP_MAXSUF)
	{
		if (suf < lp->flags)
			break;

		suf++;
		lp = lp->next;
	}

	return suf;
}

static int change_2nickx(aClient *cptr)
{
	char       newnick[NICKLEN + 1];
	int        suf;

	if ((suf = find_free_nickx()) == CHANGECP_MAXSUF + 1)
		return 0;

	snprintf(newnick, NICKLEN, "%s%u", CHANGECP_NICK, suf);

	cptr->umode &= ~UMODE_r;
#ifdef ANTI_NICK_FLOOD
	cptr->last_nick_change = NOW;
#endif
	sendto_common_channels(cptr, ":%s NICK :%s", cptr->name, newnick);
	add_history(cptr, 1);
	sendto_serv_butone(cptr, ":%s NICK %s :%ld", cptr->name, newnick,
			cptr->tsinfo);
	if (cptr->name[0])
	{
		(void) del_from_client_hash_table(cptr->name, cptr);
		hash_check_watch(cptr, RPL_LOGOFF);
	}
	set_name(cptr, newnick);
	(void) add_to_client_hash_table(newnick, cptr);
	hash_check_watch(cptr, RPL_LOGON);

	return 1;
}

int m_codepage(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aCodepage *cp;

	if (!(sptr->user && MyClient(sptr)))
		return 0;

	if (parc < 2)
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name,
			parv[0], MSG_CODEPAGE);
		return 0;
	}

	cp = findcodepage_byname(parv[1]);

	if (cp)
	{
		if (cp == sptr->codepage)
		{
			sendto_one(sptr, rpl_str(ERR_SAMECODEPAGE), me.name,
				parv[0], cp);
			return 0;
		}

		if (getstrflags(base_codepage, sptr->name) & STR_HAS8BIT)
		{
			if (!change_2nickx(sptr))
			{
				sendto_one(sptr, rpl_str(ERR_ALLNICKXBUSY),
					me.name, parv[0], CHANGECP_NICK);
				return 0;
			}
			sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Your "
				"8-bit nickname has been changed to %s due to "
				"change codepage request.", me.name,
				sptr->name, sptr->name);
		}

		change_codepage(sptr, cp);

		sendto_one(sptr, rpl_str(RPL_CODEPAGE), me.name,
			sptr->name, cp->name);
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- It is "
			"recommended to rejoin all channels you are in after "
			"changing codepage.", me.name, sptr->name);
	}
	else
		sendto_one(sptr, err_str(ERR_NOSUCHCODEPAGE), me.name,
			parv[0], parv[1]);

	return 0;
}

int m_codepages(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aCodepage  *cp;
	
	if (
		parc < 2 ||
		hunt_server(cptr, sptr, ":%s CODEPAGES %s", 1,
			parc, parv) == HUNTED_ISME
	)
	{
		cp = codepages;
		while (cp)
		{
			if (cp->flags & CODEPAGE_ACTIVE)
				sendto_one(sptr, rpl_str(RPL_CODEPAGES),
					me.name, parv[0], cp->name,
					cp->aliases ?
						cp->aliases->value.cp : "");

			cp = cp->next;
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFCODEPAGES), me.name, parv[0]);
	}

	return 0;
}

void report_codepages(aClient *cptr)
{
	aCodepage      *cp;

	cp = codepages;
	while (cp)
	{
		sendto_one(cptr, rpl_str(RPL_STATSCODEPAGE), me.name,
			cptr->name, 'B', cp->name, cp->clients, cp->maxclients,
			cp->str_flags, cp->str_ports,
			cp->aliases ? cp->aliases->value.cp : "NONE",
			cp->filename);

		cp = cp->next;
	}
}

/*
 * TODO:
 * Actually, it is better to send the SCPS message even before PASS message to
 * support 8-bit passwords and to prevent generally possible excessive
 * translation some parts of the servers joining conversation twice.
 * Most likely, it will be implemented so later, but now I'm not sure that
 * breaking the standart protocol in a such way is good.
 * -rzhe
 */
int m_scps(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aCodepage *cp;

	if (!(IsUnknown(cptr) || IsHandshake(cptr)))
		return 0;
	
	if (parc < 2 || !*parv[1])
		DupString(cptr->scps, "");
	else
	{
		DupString(cptr->scps, parv[1]);

		/*
		 * If so, the codepage will be changed to the passed in parv[1]
		 * or connection will be cancelled.
		 * -rzhe
		 */
		if (
			IsHandshake(cptr) &&
			(cp = findcodepage_byname(parv[1])) &&
			cptr->codepage != cp
		)
		{
			change_codepage(cptr, cp);

			return RETRANSL8_BUFFER;
		}
	}
	
	return 0;
}

int m_forcecp(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
	aClient        *acptr;
	char           *nick, *cpname;
	aCodepage      *cp;
	
	if (parc < 3)
	{
		sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name,
			parv[0], MSG_FORCECP);
		return 0;
	}

	if (!IsPrivileged(cptr))
	{
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	nick = parv[1];
	if (!((acptr = find_client(nick, NULL)) && IsClient(acptr)))
	{
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0],
			nick);
		return 0;
	}

	if (!(MyConnect(acptr) ? IsAnOper(sptr) : IsOper(sptr)))
	{
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	}

	cpname = parv[2];
	if (MyConnect(acptr))
	{
		if (!(cp = findcodepage_byname(cpname)))
		{
			sendto_one(sptr, err_str(ERR_NOSUCHCODEPAGE), me.name,
				parv[0], cpname);
			return 0;
		}

		if (cp == acptr->codepage)
		{
			sendto_one(sptr, rpl_str(ERR_SAMECODEPAGE), me.name,
				parv[0], cp);
			return 0;
		}

		sendto_serv_butone(NULL, ":%s GLOBOPS :%s [%s] used FORCECP "
			"on %s [%s] -- %s",
			/*
			 * As limited above, only users can be sptr for FORCECP
			 * message.
			 * -rzhe
			 */
			me.name, sptr->name, sptr->user->server, acptr->name,
			acptr->user->server, cp->name);
		send_globops("from %s: %s [%s] used FORCECP on %s [%s] -- %s",
			me.name, sptr->name, sptr->user->server, acptr->name,
			acptr->user->server, cp->name);

		sendto_one(acptr, ":%s NOTICE %s :*** Notice -- Your codepage "
			"change request received from %s [%s]", me.name,
			acptr->name, sptr->name, sptr->user->server);

		if (getstrflags(base_codepage, acptr->name) & STR_HAS8BIT)
		{
			if (!change_2nickx(acptr))
			{
				sendto_one(sptr, rpl_str(ERR_ALLNICKXBUSY),
					me.name, parv[0]);
				return 0;
			}
			sendto_one(acptr, ":%s NOTICE %s :*** Notice -- Your "
				"8-bit nickname has been changed to %s due to "
				"change codepage request.", me.name,
				acptr->name, acptr->name);
		}
		
		change_codepage(acptr, cp);
		
		sendto_one(acptr, rpl_str(RPL_CODEPAGE), me.name, acptr->name,
			cp->name);
		sendto_one(acptr, ":%s NOTICE %s :*** Notice -- It is "
			"recommended to rejoin all channels you are in after "
			"changing codepage.", me.name, acptr->name);
	}
	else
		sendto_one(acptr->from, ":%s FORCECP %s :%s", sptr->name,
			nick, cpname);

	return 0;
}

void set_name(aClient *cptr, char *name)
{
	/*
	 * Not checking for NULL pointers to not waste CPU.
	 * Just assume that this is a macro and do not call this function
	 * with cptr, cptr->name or name equal to NULL.
	 * -rzhe
	 */
	char          *dst = cptr->name;
	unsigned char  f, flags = 0, wordflags = 0, *flagstab;
	int            mixedalphawords = 0;

	if ((flagstab = base_codepage ? base_codepage->flagstab : NULL))
		while ((*dst = *name++))
		{
			f = flagstab[(cpatom)*dst];

			flags |= (*dst & 0x80) | f;

			if (f & ATOM_ALPHANUMERIC)
				wordflags |= f;
			else
				wordflags = 0;
			if ((wordflags & ATOM_MIXEDALPHA) == ATOM_MIXEDALPHA)
				mixedalphawords = 1;

			dst++;
		}
	else
		while ((*dst = *name++))
			flags |= *dst++ & 0x80;

	cptr->flags &= ~(FLAGS_HAS8BIT | FLAGS_MIXEDALPHA |
		FLAGS_MIXEDALPHAWORDS);
	if (flags & STR_HAS8BIT)
		cptr->flags |= FLAGS_HAS8BIT;
	if ((flags & STR_MIXEDALPHA) == STR_MIXEDALPHA)
		cptr->flags |= FLAGS_MIXEDALPHA;
	if (mixedalphawords)
		cptr->flags |= FLAGS_MIXEDALPHAWORDS;
}

void add_to_nickx_list(char *name)
{
	char *ch;
	Link *lp, *lpprev, *t;
	int   suf;

	if (strncasecmp(name, CHANGECP_NICK, sizeof(CHANGECP_NICK) - 1) ||
			!(*(name + sizeof(CHANGECP_NICK) - 1)))
		return;
	ch = name + sizeof(CHANGECP_NICK) - 1;
	while (*ch && isdigit((int) *ch)) ch++;
	if (*ch)
		return;

	suf = atoi(name + sizeof(CHANGECP_NICK) - 1);
	t = make_link();
	t->value.cp = name;
	t->flags = suf;
	t->next = NULL;
	
	if ((lp = lpprev = nickx))
		while (lp)
		{
			if (suf < lp->flags)
			{
				t->next = lp;
				if (lp == nickx)
					nickx = t;
				else
					lpprev->next = t;
				break;
			}
			else if (!lp->next)
			{
				lp->next = t;
				break;
			}

			lpprev = lp;
			lp = lp->next;
		}
	else
		nickx = t;
}

void del_from_nickx_list(char *name)
{
	char *ch;
	Link *lp, *lpprev, *t;

	/*
	 * TODO:
	 * Is it better to remove the paragraph below?
	 * -rzhe
	 */
	if (strncasecmp(name, CHANGECP_NICK, sizeof(CHANGECP_NICK) - 1) ||
			!(*(name + sizeof(CHANGECP_NICK) - 1)))
		return;
	ch = name + sizeof(CHANGECP_NICK) - 1;
	while (*ch && isdigit((int) *ch)) ch++;
	if (*ch)
		return;

	lp = lpprev = nickx;
	while (lp)
	{
		if (lp->value.cp == name)
		{
			if (lp == nickx)
				nickx = lp->next;
			else
				lpprev->next = lp->next;
			t = lp;
			lp = lp->next;
			free_link(t);
		}
		else
		{
			lpprev = lp;
			lp = lp->next;
		}
	}
}


#ifdef PROFILING
#define __func(type) type
#include "translate.inc"
#endif
